/**********************************************************************
 * perdition.c                                           September 1999
 * Horms                                             horms@vergenet.net
 *
 * Perdition, POP3 and IMAP4 proxy daemon
 *
 * Released under the terms of the GNU GPL
 **********************************************************************/

#include "perdition.h"

/*Use uname information here and there to idinify this system*/
struct utsname *system_uname;

/* Local and Peer address information is gloabal so perditiondb
 * libaries can access this information
 */
struct sockaddr_in *peername;
struct sockaddr_in *sockname;

/*
 * Used for opening dynamic server lookup library
 * Kept global so they can be used in signal handlers
 */
static int (*dbserver_get)(char *, char *, char **, size_t *);
static void *handle;

static void perdition_reread_handler(int sig);

/* Macro to clean things up when we jump around in the main loop*/
#define PERDITION_CLEAN_UP_MAIN \
  if(pw.pw_name!=NULL){ free(pw.pw_name); pw.pw_name=NULL; } \
  if(pw.pw_passwd!=NULL){ free(pw.pw_passwd); pw.pw_passwd=NULL;} \
  server_port_destroy(server_port); server_port=NULL; \
  destroy_token(&tag);

/* Macro to set the uid and gid */
#define PERDITION_SET_UID_AND_GID \
  if(opt.debug && geteuid()){ \
    syslog( \
      LOG_INFO,  \
      "Warning: not invoked as root, local authentication may fail" \
    ); \
  } \
  if(!geteuid() && daemon_setid(opt.username, opt.group)){ \
    if(opt.debug){ \
      syslog(LOG_WARNING, "Error setting group and userid"); \
    } \
    daemon_exit_cleanly(-1); \
  }

/**********************************************************************
 * Muriel the main funtion
 **********************************************************************/

int main (int argc, char **argv){
  struct sockaddr_in from;
  struct sockaddr_in to;
  struct passwd pw;
  unsigned char *server_ok_buf=NULL;
  unsigned char *buffer;
  server_port_t *server_port=NULL;
  protocol_t *protocol=NULL;
  token_t *tag=NULL;
  size_t server_ok_buf_size=0;
  char from_to_str[36];
  char from_str[17];
  char to_str[17];
  char *servername=NULL;
  char *port=NULL;
  int bytes_written=0;
  int bytes_read=0;
  int client_in=-1;
  int client_out=-1;
  int server=-1;
  int status;
  int round_robin_server=0;
  int rnd;

  extern struct sockaddr_in *peername;
  extern struct sockaddr_in *sockname;
  extern struct utsname *system_uname;
  extern options_t opt;

  /*Parse options*/
  options(argc, argv, OPT_FIRST_CALL);

  /*Read congif file*/
  if(opt.config_file!=NULL){
    config_file_to_opt(opt.config_file);
  }

  /*Open the dbserver_get library, if we have a library*/
  if(
    opt.map_library!=NULL && 
    getserver_openlib(
      opt.map_library,
      opt.map_library_opt,
      &handle,&dbserver_get
    )<0
  ){
    fprintf(stderr, "dlopen of \"%s\" failed\n", opt.map_library);
    usage(-1);
    daemon_exit_cleanly(-1);
  }

  /*Set signal handlers*/
  signal(SIGHUP,    (void(*)(int))perdition_reread_handler);
  signal(SIGINT,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGQUIT,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGILL,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGTRAP,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGIOT,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGBUS,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGFPE,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGUSR1,   (void(*)(int))daemon_noop_handler);
  signal(SIGSEGV,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGUSR2,   (void(*)(int))daemon_noop_handler);
  signal(SIGPIPE,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGALRM,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGTERM,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGURG,    (void(*)(int))daemon_exit_cleanly);
  signal(SIGXCPU,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGXFSZ,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGVTALRM, (void(*)(int))daemon_exit_cleanly);
  signal(SIGPROF,   (void(*)(int))daemon_exit_cleanly);
  signal(SIGWINCH,  (void(*)(int))daemon_exit_cleanly);
  signal(SIGIO,     (void(*)(int))daemon_exit_cleanly);

  /*Close file descriptors and detactch process from shell as necessary*/
  if(opt.inetd_mode){
    daemon_inetd_process();
  }
  else{
    daemon_process();
  }

  /*Open syslog, now we have detached*/
  closelog();
  openlog(LOG_IDENT, LOG_OPTION, LOG_FACILITY);

  /*Seed the uname structure*/
  if((system_uname=(struct utsname *)malloc(sizeof(struct utsname)))==NULL){
    syslog(LOG_WARNING, "main: malloc system_uname: %s", strerror(errno));
    daemon_exit_cleanly(-1);
  }
  if(uname(system_uname)<0){
    syslog(LOG_WARNING, "main: uname");
    daemon_exit_cleanly(-1);
  }

  /*Set up protocol structure*/
  if((protocol=protocol_initialise(opt.protocol, protocol))==NULL){
    syslog(LOG_WARNING, "main: protocol_initialise");
    daemon_exit_cleanly(-1);
  }

  /*Set listen and outgoing port now the protocol structure is accessable*/
  if((opt.listen_port=(*(protocol->port))(opt.listen_port))==NULL){
    syslog(LOG_WARNING, "main: protocol->port 1");
    daemon_exit_cleanly(-1);
  }
  if((opt.outgoing_port=(*(protocol->port))(opt.outgoing_port))==NULL){
    syslog(LOG_WARNING, "main: protocol->port 2");
    daemon_exit_cleanly(-1);
  }

  /*
   * Log the options we will be running with.
   * If we are in inetd mode then only do this if debuging is turned on,
   * else dbuging is a bit too verbose.
   */
  if((!opt.inetd_mode) || opt.debug){
    if(log_options()){
      syslog(LOG_WARNING, "main: log_options");
      daemon_exit_cleanly(-1);
    }
  }

  /*If we are using the server's ok line then allocate a buffer to store it*/
  if(opt.server_ok_line){
    if((server_ok_buf=(unsigned char *)malloc(
      sizeof(unsigned char)*MAX_LINE_LENGTH
    ))==NULL){
      syslog(LOG_WARNING, "main: malloc server_ok_buf: %s",  strerror(errno));
      daemon_exit_cleanly(-1);
    }
  }

  /*Open incoming socket as required*/
  if(opt.inetd_mode){
    client_in=0;
    client_out=1;
    *from_to_str='\0';
    *to_str='\0';
    *from_str='\0';
    peername=NULL;
    sockname=NULL;
  }
  else{
    if((client_in=tcp_socket_server_connect(
      opt.listen_port, 
      opt.bind_address,
      opt.connection_limit, 
      &from,
      &to
    ))<0){
      syslog(
        LOG_WARNING, 
        "main: tcp_socket_server_connect: %s",
        strerror(errno)
      );
      daemon_exit_cleanly(-1);
    }
    client_out=client_in;
    snprintf(from_str, 17, "%s", inet_ntoa(from.sin_addr));
    snprintf(to_str,   17, "%s", inet_ntoa(to.sin_addr));
    snprintf(from_to_str, 36, "%s->%s ", from_str, to_str);
    peername=&from;
    sockname=&to;
  }

  /*Become someone else*/
  /*NB: We do this later if we are going to authenticate locally*/
  if(!opt.authenticate_in){
    PERDITION_SET_UID_AND_GID;
  }

  /*Seed rand*/
  srand(time(NULL)*getpid());
  rnd=rand();

  /*Speak to our client*/
  if(greeting(client_out, protocol, GREETING_ADD_NODENAME)){
    syslog(LOG_WARNING, "main: greeting");
    daemon_exit_cleanly(-1);
  }

  pw.pw_name=NULL;
  pw.pw_passwd=NULL;
  /* Authenticate the user*/
  for(;;){
    /*Read the USER and PASS lines from the client */
    if((status=(*(protocol->in_get_pw))(client_in, client_out, &pw, &tag))<0){
      syslog(LOG_WARNING, "main: protocol->in_get_pw");
      daemon_exit_cleanly(-1);
    }
    else if(status>0){
      if(opt.debug){
        syslog(
          LOG_INFO, 
          "Closing NULL session: %susername=%s", 
          from_to_str,
          str_null_safe(pw.pw_name)
        );
      }
      daemon_exit_cleanly(0);
    }

    /*Read the server from the map, if we have a map*/
    if(
      opt.map_library!=NULL &&
      (server_port=getserver(pw.pw_name, dbserver_get))!=NULL
    ){
      char *host;
      port=server_port_get_port(server_port);
      servername=server_port_get_servername(server_port);
      if((host=strstr(servername, opt.domain_delimiter)) != NULL){
        char *temp1;
        char *temp2;
        /* The Username */
        temp1 = strn_to_str(servername,(size_t)(host-servername));

        /* The Host */
        host++;
        temp2 = strn_to_str(host,strlen(host));
        
        free(pw.pw_name);
        pw.pw_name = temp1;
        free(servername);
        servername=temp2;
	if(opt.debug){
          syslog(
	    LOG_INFO, 
	    "main: transresult:%s,%s",
	    pw.pw_name,
	    servername
          );
	}
      }
    }

    /*Use the default server if we have one and the servername is not set*/
    if((servername==NULL || round_robin_server) && opt.outgoing_server!=NULL){
      round_robin_server=1;
      rnd=(rnd+1)%dynamic_array_get_count(opt.outgoing_server);
      server_port=dynamic_array_get_element(opt.outgoing_server,rnd);
      servername=server_port_get_servername(server_port);
      port=server_port_get_port(server_port);
    }

    /*Use the default port if the port is not set*/
    if(port==NULL){
      port=opt.outgoing_port;
    }

    /*Log the session*/
    syslog(
      LOG_INFO, 
      "Connect: %suser=%s server=%s port=%s", 
      from_to_str,
      str_null_safe(pw.pw_name),
      str_null_safe(servername),
      str_null_safe(port)
    );

    /*Try again ig we didn't get anything useful*/
    if(servername==NULL){
      sleep(PERDITION_ERR_SLEEP);
      if((*(protocol->write))(
        client_out, 
        NULL_FLAG,
	tag,
	protocol->type[PROTOCOL_ERR], 
	"Could not determine server"
      )<0){
        syslog(LOG_WARNING, "main: protocol->write");
        daemon_exit_cleanly(-1);
      }
      PERDITION_CLEAN_UP_MAIN;
      continue;
    }

    if(opt.authenticate_in){
      /*Authenticate user locally - Not really necessary*/
      if(opt.debug && geteuid()){ \
        syslog(
          LOG_INFO,
          "Warning: not invoked as root, local authentication may fail"
        );
      }

      if((status=protocol->in_authenticate(&pw, client_out, tag))==0){
        syslog(LOG_INFO, "Local authentication failure");
        PERDITION_CLEAN_UP_MAIN;
        continue;
      }
      else if(status<0){
        syslog(LOG_WARNING, "main: pop3_in_authenticate");
        daemon_exit_cleanly(-1);
      }
    
      /*
       * If local authentcation is used then now is the time to
       * Become someone else
       */
       PERDITION_SET_UID_AND_GID;
    }

    /* Talk to the real pop server for the client*/
    if((server=tcp_socket_client_open(servername, port))<0){
      syslog(LOG_INFO, "Could not connect to server");
      sleep(PERDITION_ERR_SLEEP);
      if((*(protocol->write))(
        client_out,
        NULL_FLAG,
	tag,
	protocol->type[PROTOCOL_ERR], 
	"Could not connect to server"
      )<0){
        syslog(LOG_WARNING, "main: protocol->write");
        daemon_exit_cleanly(-1);
      }
      PERDITION_CLEAN_UP_MAIN;
      continue;
    }

    /*Authenticate the user with the pop server*/
    if(opt.server_ok_line){
      server_ok_buf_size=MAX_LINE_LENGTH-1;
    }
    status = (*(protocol->out_authenticate))(
      server, 
      server, 
      &pw, 
      tag,
      protocol,
      server_ok_buf,
      &server_ok_buf_size
    );

    if(status==0){
      sleep(PERDITION_ERR_SLEEP);
      syslog(LOG_INFO, "Fail reauthentication for user %s", pw.pw_name);
      quit(server, server, protocol);
      if(close(server)){
        syslog(LOG_WARNING, "main: close(server) 2");
	daemon_exit_cleanly(-1);
      }
      if(protocol->write(
        client_out, 
        NULL_FLAG,
        NULL, 
        protocol->type[PROTOCOL_NO], 
        "Re-Authentecation Failure"
      )<0){
        syslog(LOG_WARNING, "main: protocol->write");
        daemon_exit_cleanly(-1);
      }
      PERDITION_CLEAN_UP_MAIN;
      continue;
    }
    else if(status<0){
      syslog(LOG_WARNING, "main: protocol->out_authenticate %d", status);
      daemon_exit_cleanly(-1);
    }

    /*If we get this far, dance for joy with lmf*/
    if(opt.server_ok_line){
      *(server_ok_buf+server_ok_buf_size)='\0';
      buffer=server_ok_buf;
      if(protocol->write(
        client_out, 
        WRITE_STR_NO_CLLF,
        tag, 
        NULL,
        server_ok_buf
      )<0){
        syslog(LOG_WARNING, "main: protocol->write");
        daemon_exit_cleanly(-1);
      }
    }
    else{
      if(protocol->write(
        client_out, 
        NULL_FLAG,
        tag, 
        protocol->type[PROTOCOL_OK],
        "You are so in"
      )<0){
        syslog(LOG_WARNING, "main: protocol->write");
        daemon_exit_cleanly(-1);
      }
    }

    break;
  }

  if(opt.server_ok_line){
     free(server_ok_buf);
  }

  /*We need a buffer for reads and writes to the server*/
  if((buffer=(unsigned char *)malloc(BUFFER_SIZE*sizeof(unsigned char)))==NULL){
    syslog(LOG_WARNING, "main: malloc: %s, Exiting", strerror(errno));
    daemon_exit_cleanly(-1);
  }

  /*Let the client talk to the real server*/
  if(tcp_socket_pipe(
    server,
    server,
    client_in,
    client_out,
    buffer,
    BUFFER_SIZE,
    opt.timeout,
    &bytes_written,
    &bytes_read
  )<0){
    syslog(LOG_WARNING, "main: tcp_socket_pipe: %s", strerror(errno));
    daemon_exit_cleanly(-1);
  }

  /*Time to leave*/
  syslog(
    LOG_INFO, 
    "Closing: %suser=%s %d %d", 
    from_to_str,
    pw.pw_name,
    bytes_read,
    bytes_written
  );
  getserver_closelib(handle);
  daemon_exit_cleanly(0);

  PERDITION_CLEAN_UP_MAIN;
  /*Here so compilers won't barf*/
  return(0);
}


/**********************************************************************
 * perdition_reread_handler
 * A signal handler that closes and opens the map libary,
 * reinitialising as necessary.
 * pre: sig: signal recieved by the process
 * post: signal handler reset for signal
 *       map library closed and opened
 *       This may casuse shutdown and initialisation of map to
 *       take place if appropriate symbols are defined in
 *       the library. See getserver.c for details.
 **********************************************************************/

static void perdition_reread_handler(int sig){
  extern options_t opt;

  getserver_closelib(handle);
  if(
    opt.map_library!=NULL && 
    getserver_openlib(
      opt.map_library, 
      opt.map_library_opt, 
      &handle, 
      &dbserver_get
    )<0
  ){
    syslog(LOG_WARNING, "main: getserver_openlib");
    daemon_exit_cleanly(-1);
  }

  signal(sig, (void(*)(int))perdition_reread_handler);
}
