/**********************************************************************
 * pop3_write.c                                          September 1999
 * Horms                                             horms@vergenet.net
 *
 * Write POP3 protocol commands and responses
 *
 * Licenced under the terms of the GNU GPL
 *
 **********************************************************************/

#include "pop3_write.h"
#include "options.h"

/**********************************************************************
 * pop3_write
 * Write a message of the form [<type> ]<string>
 * Pre: fd: file descriptor to write to
 *      flag: flag to pass to write_str, as per str.h
 *      tag: ignored
 *      type: type of message, POP3_OK or POP3_ERR
 *            if NULL then only string is written
 *      string: mesage to display
 * Return 0 on success
 *        -1 otherwise
 **********************************************************************/
      
int pop3_write(
  const int fd, 
  const flag_t flag,
  const token_t *tag,
  const char *type, 
  const char *string
){
  extern options_t opt;

  if(type==NULL){
    if(write_str(fd, flag, 1, string)<0){
      if(opt.debug){
        syslog(LOG_WARNING, "pop3_write: write_str");
      }
      return(-1);
    }   
  }
  else {
    if(write_str(fd, flag, 3, type, " ", string)<0){
      if(opt.debug){
        syslog(LOG_WARNING, "pop3_write: write_str");
      }
      return(-1);
    }   
  }
  return(0);
}   
